vim:syntax=text

GPL, copyright Ian Lynagh <igloo@earth.li> 2001

> module IA (Automaton_State, init_states, step_automaton, no_such_state) where

> import Maybe
> import Tokens
> import FiniteMap
> import List

> data StateChange = Jump [Integer] 
>                  | Long_Jump Integer Integer
>                  | SLiteral (Either Token Func) Integer
>                  | End

> data Func = Func String (Token -> Bool)

> instance Eq Func where
>   Func x _ == Func y _ = x == y

> instance Ord Func where
>   Func x _ <= Func y _ = x <= y

> type LocalState = (Integer, Integer) -- The position in a line of definition
> type State = [LocalState] -- The stack of local states this state is in
> type Automaton_State = [State] -- The current list of states we could be in

> init_states :: Automaton_State
> init_states = [[( 0,  0)]]

> no_such_state :: Automaton_State -> Bool
> no_such_state = (==) []

> step_automaton :: Automaton_State -> Token -> Automaton_State
> step_automaton ss t = [ls | (e, ls) <- closure ss, can_step e (un_im t)]

> un_im :: Token -> Token
> un_im (ImplicitSpecial s) = Special s
> un_im t = t

> can_step :: Either Token Func -> Token -> Bool
> can_step (Left s) t = s == t
> can_step (Right (Func _ f)) t = f t

> closure :: [State] -> [(Either Token Func, State)]
> closure = remove_dupes . sort . do_all_nulls

> remove_dupes :: (Eq a) => [a] -> [a]
> remove_dupes (a:b:xs)
>  | a == b    =   remove_dupes (b:xs)
>  | otherwise = a:remove_dupes (b:xs)
> remove_dupes xs = xs

> do_all_nulls :: [State] -> [(Either Token Func, State)]
> do_all_nulls [] = []
> do_all_nulls ([]:qs) = do_all_nulls qs
> do_all_nulls ((qh:qt):qs)
>  = case lookupFM nulls_fm qh of
>        Nothing -> error "do_all_nulls: Can't happen (Nothing)"
>        Just x -> case x of
>             Long_Jump i j -> do_all_nulls (((i, 0):(h, j):qt):qs)
>             SLiteral e i -> (e, (h, i):qt):do_all_nulls qs
>             End -> do_all_nulls (qt:qs)
>             Jump xs -> do_all_nulls (tailed_xs ++ qs)
>                 where tailed_xs = map ((: qt) . (,) h) xs
>   where h = fst qh

> nulls_fm :: FiniteMap LocalState StateChange
> nulls_fm = listToFM nulls

> nulls :: [(LocalState, StateChange)]
> nulls =[((0,7),End),((0,0),Jump [1,6]),((0,1),SLiteral (Left (ReservedID "module")) 2),((0,2),SLiteral (Right (Func "is_ConID" (is_ConID))) 3),((0,3),Jump [4,5]),((0,4),Long_Jump 3 5),((0,5),SLiteral (Left (ReservedID "where")) 6),((0,6),Long_Jump 1 7),((1,12),End),((1,0),Jump [1,6,9]),((1,1),SLiteral (Left (Special '{')) 2),((1,2),Long_Jump 2 3),((1,3),SLiteral (Left (Special ';')) 4),((1,4),Long_Jump 9 5),((1,5),SLiteral (Left (Special '}')) 12),((1,6),SLiteral (Left (Special '{')) 7),((1,7),Long_Jump 2 8),((1,8),SLiteral (Left (Special '}')) 12),((1,9),SLiteral (Left (Special '{')) 10),((1,10),Long_Jump 9 11),((1,11),SLiteral (Left (Special '}')) 12),((2,3),End),((2,1),Jump [2,3]),((2,0),Long_Jump 5 1),((2,2),SLiteral (Left (Special ';')) 0),((3,8),End),((3,0),SLiteral (Left (Special '(')) 1),((3,1),Jump [2,5]),((3,3),Jump [4,5]),((3,2),Long_Jump 4 3),((3,4),SLiteral (Left (Special ',')) 2),((3,5),Jump [6,7]),((3,6),SLiteral (Left (Special ',')) 7),((3,7),SLiteral (Left (Special ')')) 8),((4,16),End),((4,0),Jump [1,2,14]),((4,1),Long_Jump 60 16),((4,2),SLiteral (Right (Func "is_QConID" (is_QConID))) 3),((4,3),Jump [4,16]),((4,4),SLiteral (Left (Special '(')) 5),((4,5),Jump [6,7]),((4,6),SLiteral (Left (ReservedOp "..")) 13),((4,7),Jump [8,13]),((4,11),Jump [12,13]),((4,8),Jump [9,10]),((4,9),Long_Jump 8 11),((4,10),Long_Jump 60 11),((4,12),SLiteral (Left (Special ',')) 8),((4,13),SLiteral (Left (Special ')')) 16),((4,14),SLiteral (Left (ReservedID "module")) 15),((4,15),SLiteral (Right (Func "is_ConID" (is_ConID))) 16),((5,11),End),((5,0),Jump [1,10]),((5,1),SLiteral (Left (ReservedID "import")) 2),((5,2),Jump [3,4]),((5,3),SLiteral (Right (Func "is_VarID_name \"qualified\"" (is_VarID_name "qualified"))) 4),((5,4),SLiteral (Right (Func "is_ConID" (is_ConID))) 5),((5,5),Jump [6,8]),((5,6),SLiteral (Right (Func "is_VarID_name \"as\"" (is_VarID_name "as"))) 7),((5,7),SLiteral (Right (Func "is_ConID" (is_ConID))) 8),((5,8),Jump [9,11]),((5,9),Long_Jump 6 11),((5,10),Jump [11]),((6,10),End),((6,0),Jump [1,2]),((6,1),SLiteral (Right (Func "is_VarID_name \"hiding\"" (is_VarID_name "hiding"))) 2),((6,2),SLiteral (Left (Special '(')) 3),((6,3),Jump [4,7]),((6,5),Jump [6,7]),((6,4),Long_Jump 7 5),((6,6),SLiteral (Left (Special ',')) 4),((6,7),Jump [8,9]),((6,8),SLiteral (Left (Special ',')) 9),((6,9),SLiteral (Left (Special ')')) 10),((7,13),End),((7,0),Jump [1,2]),((7,1),Long_Jump 59 13),((7,2),SLiteral (Right (Func "is_ConID" (is_ConID))) 3),((7,3),Jump [4,13]),((7,4),Jump [5,8]),((7,5),SLiteral (Left (Special '(')) 6),((7,6),SLiteral (Left (ReservedOp "..")) 7),((7,7),SLiteral (Left (Special ')')) 13),((7,8),SLiteral (Left (Special '(')) 9),((7,10),Jump [11,12]),((7,9),Long_Jump 8 10),((7,11),SLiteral (Left (Special ',')) 9),((7,12),SLiteral (Left (Special ')')) 13),((8,3),End),((8,0),Jump [1,2]),((8,1),Long_Jump 59 3),((8,2),Long_Jump 61 3),((9,4),End),((9,0),Jump [1,4]),((9,2),Jump [3,4]),((9,1),Long_Jump 10 2),((9,3),SLiteral (Left (Special ';')) 1),((10,49),End),((10,0),Jump [1,5,14,23,32,41,48]),((10,1),SLiteral (Left (ReservedID "type")) 2),((10,2),Long_Jump 29 3),((10,3),SLiteral (Left (ReservedOp "=")) 4),((10,4),Long_Jump 21 49),((10,5),SLiteral (Left (ReservedID "data")) 6),((10,6),Jump [7,9]),((10,7),Long_Jump 25 8),((10,8),SLiteral (Left (ReservedOp "=>")) 9),((10,9),Long_Jump 29 10),((10,10),SLiteral (Left (ReservedOp "=")) 11),((10,11),Long_Jump 30 12),((10,12),Jump [13,49]),((10,13),Long_Jump 34 49),((10,14),SLiteral (Left (ReservedID "newtype")) 15),((10,15),Jump [16,18]),((10,16),Long_Jump 25 17),((10,17),SLiteral (Left (ReservedOp "=>")) 18),((10,18),Long_Jump 29 19),((10,19),SLiteral (Left (ReservedOp "=")) 20),((10,20),Long_Jump 32 21),((10,21),Jump [22,49]),((10,22),Long_Jump 34 49),((10,23),SLiteral (Left (ReservedID "class")) 24),((10,24),Jump [25,27]),((10,25),Long_Jump 27 26),((10,26),SLiteral (Left (ReservedOp "=>")) 27),((10,27),SLiteral (Right (Func "is_ConID" (is_ConID))) 28),((10,28),SLiteral (Right (Func "is_VarID" (is_VarID))) 29),((10,29),Jump [30,49]),((10,30),SLiteral (Left (ReservedID "where")) 31),((10,31),Long_Jump 13 49),((10,32),SLiteral (Left (ReservedID "instance")) 33),((10,33),Jump [34,36]),((10,34),Long_Jump 27 35),((10,35),SLiteral (Left (ReservedOp "=>")) 36),((10,36),SLiteral (Right (Func "is_QConID" (is_QConID))) 37),((10,37),Long_Jump 36 38),((10,38),Jump [39,49]),((10,39),SLiteral (Left (ReservedID "where")) 40),((10,40),Long_Jump 15 49),((10,41),SLiteral (Left (ReservedID "default")) 42),((10,42),SLiteral (Left (Special '(')) 43),((10,43),Jump [44,47]),((10,45),Jump [46,47]),((10,44),Long_Jump 21 45),((10,46),SLiteral (Left (Special ',')) 44),((10,47),SLiteral (Left (Special ')')) 49),((10,48),Long_Jump 12 49),((11,6),End),((11,0),SLiteral (Left (Special '{')) 1),((11,1),Jump [2,5]),((11,3),Jump [4,5]),((11,2),Long_Jump 12 3),((11,4),SLiteral (Left (Special ';')) 2),((11,5),SLiteral (Left (Special '}')) 6),((12,6),End),((12,0),Jump [1,2]),((12,1),Long_Jump 17 6),((12,2),Jump [3,4]),((12,3),Long_Jump 37 5),((12,4),Long_Jump 54 5),((12,5),Long_Jump 38 6),((13,6),End),((13,0),SLiteral (Left (Special '{')) 1),((13,1),Jump [2,5]),((13,3),Jump [4,5]),((13,2),Long_Jump 14 3),((13,4),SLiteral (Left (Special ';')) 2),((13,5),SLiteral (Left (Special '}')) 6),((14,6),End),((14,0),Jump [1,2]),((14,1),Long_Jump 17 6),((14,2),Jump [3,4]),((14,3),Long_Jump 37 5),((14,4),Long_Jump 59 5),((14,5),Long_Jump 38 6),((15,6),End),((15,0),SLiteral (Left (Special '{')) 1),((15,1),Jump [2,5]),((15,3),Jump [4,5]),((15,2),Long_Jump 16 3),((15,4),SLiteral (Left (Special ';')) 2),((15,5),SLiteral (Left (Special '}')) 6),((16,6),End),((16,0),Jump [1,5]),((16,1),Jump [2,3]),((16,2),Long_Jump 37 4),((16,3),Long_Jump 59 4),((16,4),Long_Jump 38 6),((16,5),Jump [6]),((17,12),End),((17,0),Jump [1,7,11]),((17,1),Long_Jump 19 2),((17,2),SLiteral (Left (ReservedOp "::")) 3),((17,3),Jump [4,6]),((17,4),Long_Jump 25 5),((17,5),SLiteral (Left (ReservedOp "=>")) 6),((17,6),Long_Jump 21 12),((17,7),Long_Jump 20 8),((17,8),Jump [9,10]),((17,9),SLiteral (Right (Func "is_integer" (is_integer))) 10),((17,10),Long_Jump 18 12),((17,11),Jump [12]),((18,3),End),((18,1),Jump [2,3]),((18,0),Long_Jump 67 1),((18,2),SLiteral (Left (Special ',')) 0),((19,3),End),((19,1),Jump [2,3]),((19,0),Long_Jump 59 1),((19,2),SLiteral (Left (Special ',')) 0),((20,4),End),((20,0),Jump [1,2,3]),((20,1),SLiteral (Left (ReservedID "infixl")) 4),((20,2),SLiteral (Left (ReservedID "infixr")) 4),((20,3),SLiteral (Left (ReservedID "infix")) 4),((21,4),End),((21,0),Long_Jump 22 1),((21,1),Jump [2,4]),((21,2),SLiteral (Left (ReservedOp "->")) 3),((21,3),Long_Jump 21 4),((22,2),End),((22,1),Jump [0,2]),((22,0),Long_Jump 23 1),((23,11),End),((23,0),Jump [1,2,3,8]),((23,1),Long_Jump 24 11),((23,2),SLiteral (Right (Func "is_VarID" (is_VarID))) 11),((23,3),SLiteral (Left (Special '(')) 4),((23,5),Jump [6,7]),((23,4),Long_Jump 21 5),((23,6),SLiteral (Left (Special ',')) 4),((23,7),SLiteral (Left (Special ')')) 11),((23,8),SLiteral (Left (Special '[')) 9),((23,9),Long_Jump 21 10),((23,10),SLiteral (Left (Special ']')) 11),((24,11),End),((24,0),Jump [1,2,9]),((24,1),SLiteral (Right (Func "is_QConID" (is_QConID))) 11),((24,2),SLiteral (Left (Special '(')) 3),((24,3),Jump [4,7]),((24,4),Jump [5,8]),((24,6),Jump [5,8]),((24,5),SLiteral (Left (Special ',')) 6),((24,7),SLiteral (Left (ReservedOp "->")) 8),((24,8),SLiteral (Left (Special ')')) 11),((24,9),SLiteral (Left (Special '[')) 10),((24,10),SLiteral (Left (Special ']')) 11),((25,8),End),((25,0),Jump [1,2]),((25,1),Long_Jump 26 8),((25,2),SLiteral (Left (Special '(')) 3),((25,3),Jump [4,7]),((25,5),Jump [6,7]),((25,4),Long_Jump 26 5),((25,6),SLiteral (Left (Special ',')) 4),((25,7),SLiteral (Left (Special ')')) 8),((26,8),End),((26,0),SLiteral (Right (Func "is_QConID" (is_QConID))) 1),((26,1),Jump [2,3]),((26,2),SLiteral (Right (Func "is_VarID" (is_VarID))) 8),((26,3),SLiteral (Left (Special '(')) 4),((26,4),SLiteral (Right (Func "is_VarID" (is_VarID))) 5),((26,6),Jump [5,7]),((26,5),Long_Jump 23 6),((26,7),SLiteral (Left (Special ')')) 8),((27,8),End),((27,0),Jump [1,2]),((27,1),Long_Jump 28 8),((27,2),SLiteral (Left (Special '(')) 3),((27,3),Jump [4,7]),((27,5),Jump [6,7]),((27,4),Long_Jump 28 5),((27,6),SLiteral (Left (Special ',')) 4),((27,7),SLiteral (Left (Special ')')) 8),((28,2),End),((28,0),SLiteral (Right (Func "is_QConID" (is_QConID))) 1),((28,1),SLiteral (Right (Func "is_VarID" (is_VarID))) 2),((29,4),End),((29,0),SLiteral (Right (Func "is_ConID" (is_ConID))) 1),((29,1),Jump [2,4]),((29,3),Jump [2,4]),((29,2),SLiteral (Right (Func "is_VarID" (is_VarID))) 3),((30,3),End),((30,1),Jump [2,3]),((30,0),Long_Jump 31 1),((30,2),SLiteral (Left (ReservedOp "|")) 0),((31,23),End),((31,0),Jump [1,14]),((31,1),Long_Jump 61 2),((31,2),Jump [3,8]),((31,3),Jump [4,23]),((31,7),Jump [4,23]),((31,4),Jump [5,6]),((31,5),SLiteral (Right (Func "is_bang" (is_bang))) 6),((31,6),Long_Jump 23 7),((31,8),SLiteral (Left (Special '{')) 9),((31,9),Jump [10,13]),((31,11),Jump [12,13]),((31,10),Long_Jump 33 11),((31,12),SLiteral (Left (Special ',')) 10),((31,13),SLiteral (Left (Special '}')) 23),((31,14),Jump [15,16]),((31,15),Long_Jump 22 18),((31,16),SLiteral (Right (Func "is_bang" (is_bang))) 17),((31,17),Long_Jump 23 18),((31,18),Long_Jump 65 19),((31,19),Jump [20,21]),((31,20),Long_Jump 22 23),((31,21),SLiteral (Right (Func "is_bang" (is_bang))) 22),((31,22),Long_Jump 23 23),((32,8),End),((32,0),Long_Jump 61 1),((32,1),Jump [2,3]),((32,2),Long_Jump 23 8),((32,3),SLiteral (Left (Special '{')) 4),((32,4),Long_Jump 59 5),((32,5),SLiteral (Left (ReservedOp "::")) 6),((32,6),Long_Jump 21 7),((32,7),SLiteral (Left (Special '}')) 8),((33,6),End),((33,0),Long_Jump 19 1),((33,1),SLiteral (Left (ReservedOp "::")) 2),((33,2),Jump [3,4]),((33,3),Long_Jump 21 6),((33,4),SLiteral (Right (Func "is_bang" (is_bang))) 5),((33,5),Long_Jump 23 6),((34,9),End),((34,0),SLiteral (Left (ReservedID "deriving")) 1),((34,1),Jump [2,3]),((34,2),Long_Jump 35 9),((34,3),SLiteral (Left (Special '(')) 4),((34,4),Jump [5,8]),((34,6),Jump [7,8]),((34,5),Long_Jump 35 6),((34,7),SLiteral (Left (Special ',')) 5),((34,8),SLiteral (Left (Special ')')) 9),((35,1),End),((35,0),SLiteral (Right (Func "is_QConID" (is_QConID))) 1),((36,19),End),((36,0),Jump [1,2,16]),((36,1),Long_Jump 24 19),((36,2),SLiteral (Left (Special '(')) 3),((36,3),Jump [4,8]),((36,4),Long_Jump 24 5),((36,5),Jump [6,15]),((36,7),Jump [6,15]),((36,6),SLiteral (Right (Func "is_VarID" (is_VarID))) 7),((36,8),SLiteral (Right (Func "is_VarID" (is_VarID))) 9),((36,9),Jump [10,13]),((36,11),Jump [12,15]),((36,10),SLiteral (Left (Special ',')) 11),((36,12),SLiteral (Right (Func "is_VarID" (is_VarID))) 10),((36,13),SLiteral (Left (ReservedOp "->")) 14),((36,14),SLiteral (Right (Func "is_VarID" (is_VarID))) 15),((36,15),SLiteral (Left (Special ')')) 19),((36,16),SLiteral (Left (Special '[')) 17),((36,17),SLiteral (Right (Func "is_VarID" (is_VarID))) 18),((36,18),SLiteral (Left (Special ']')) 19),((37,12),End),((37,0),Jump [1,4,7]),((37,1),Long_Jump 59 2),((37,3),Jump [2,12]),((37,2),Long_Jump 56 3),((37,4),Long_Jump 54 5),((37,5),Long_Jump 63 6),((37,6),Long_Jump 54 12),((37,7),SLiteral (Left (Special '(')) 8),((37,8),Long_Jump 37 9),((37,9),SLiteral (Left (Special ')')) 10),((37,11),Jump [10,12]),((37,10),Long_Jump 56 11),((38,7),End),((38,0),Jump [1,3]),((38,1),SLiteral (Left (ReservedOp "=")) 2),((38,2),Long_Jump 41 4),((38,3),Long_Jump 39 4),((38,4),Jump [5,7]),((38,5),SLiteral (Left (ReservedID "where")) 6),((38,6),Long_Jump 11 7),((39,5),End),((39,0),Long_Jump 40 1),((39,1),SLiteral (Left (ReservedOp "=")) 2),((39,2),Long_Jump 41 3),((39,3),Jump [4,5]),((39,4),Long_Jump 39 5),((40,2),End),((40,0),SLiteral (Left (ReservedOp "|")) 1),((40,1),Long_Jump 42 2),((41,7),End),((41,0),Long_Jump 42 1),((41,1),Jump [2,7]),((41,2),SLiteral (Left (ReservedOp "::")) 3),((41,3),Jump [4,6]),((41,4),Long_Jump 25 5),((41,5),SLiteral (Left (ReservedOp "=>")) 6),((41,6),Long_Jump 21 7),((42,5),End),((42,3),Jump [4,5]),((42,0),Jump [1,2]),((42,1),SLiteral (Right (Func "is_negative" (is_negative))) 2),((42,2),Long_Jump 43 3),((42,4),Long_Jump 68 0),((43,27),End),((43,0),Jump [1,6,10,16,22,26]),((43,1),SLiteral (Left (ReservedOp "\\")) 2),((43,3),Jump [2,4]),((43,2),Long_Jump 56 3),((43,4),SLiteral (Left (ReservedOp "->")) 5),((43,5),Long_Jump 41 27),((43,6),SLiteral (Left (ReservedID "let")) 7),((43,7),Long_Jump 11 8),((43,8),SLiteral (Left (ReservedID "in")) 9),((43,9),Long_Jump 41 27),((43,10),SLiteral (Left (ReservedID "if")) 11),((43,11),Long_Jump 41 12),((43,12),SLiteral (Left (ReservedID "then")) 13),((43,13),Long_Jump 41 14),((43,14),SLiteral (Left (ReservedID "else")) 15),((43,15),Long_Jump 41 27),((43,16),SLiteral (Left (ReservedID "case")) 17),((43,17),Long_Jump 41 18),((43,18),SLiteral (Left (ReservedID "of")) 19),((43,19),SLiteral (Left (Special '{')) 20),((43,20),Long_Jump 47 21),((43,21),SLiteral (Left (Special '}')) 27),((43,22),SLiteral (Left (ReservedID "do")) 23),((43,23),SLiteral (Left (Special '{')) 24),((43,24),Long_Jump 50 25),((43,25),SLiteral (Left (Special '}')) 27),((43,26),Long_Jump 44 27),((44,2),End),((44,1),Jump [0,2]),((44,0),Long_Jump 45 1),((45,61),End),((45,0),Jump [1,8,9,10,13,18,24,33,40,44,48]),((45,1),Long_Jump 60 2),((45,2),Jump [3,61]),((45,3),SLiteral (Left (Special '{')) 4),((45,5),Jump [6,7]),((45,4),Long_Jump 52 5),((45,6),SLiteral (Left (Special ',')) 4),((45,7),SLiteral (Left (Special '}')) 61),((45,8),Long_Jump 58 61),((45,9),SLiteral (Right (Func "is_literal" (is_literal))) 61),((45,10),SLiteral (Left (Special '(')) 11),((45,11),Long_Jump 41 12),((45,12),SLiteral (Left (Special ')')) 61),((45,13),SLiteral (Left (Special '(')) 14),((45,15),Jump [16,17]),((45,14),Long_Jump 41 15),((45,16),SLiteral (Left (Special ',')) 14),((45,17),SLiteral (Left (Special ')')) 61),((45,18),SLiteral (Left (Special '[')) 19),((45,19),Jump [20,23]),((45,21),Jump [22,23]),((45,20),Long_Jump 41 21),((45,22),SLiteral (Left (Special ',')) 20),((45,23),SLiteral (Left (Special ']')) 61),((45,24),SLiteral (Left (Special '[')) 25),((45,25),Long_Jump 41 26),((45,26),Jump [27,29]),((45,27),SLiteral (Left (Special ',')) 28),((45,28),Long_Jump 41 29),((45,29),SLiteral (Left (ReservedOp "..")) 30),((45,30),Jump [31,32]),((45,31),Long_Jump 41 32),((45,32),SLiteral (Left (Special ']')) 61),((45,33),SLiteral (Left (Special '[')) 34),((45,34),Long_Jump 41 35),((45,35),SLiteral (Left (ReservedOp "|")) 36),((45,37),Jump [38,39]),((45,36),Long_Jump 46 37),((45,38),SLiteral (Left (Special ',')) 36),((45,39),SLiteral (Left (Special ']')) 61),((45,40),SLiteral (Left (Special '(')) 41),((45,41),Long_Jump 42 42),((45,42),Long_Jump 68 43),((45,43),SLiteral (Left (Special ')')) 61),((45,44),SLiteral (Left (Special '(')) 45),((45,45),Long_Jump 68 46),((45,46),Long_Jump 42 47),((45,47),SLiteral (Left (Special ')')) 61),((45,48),Long_Jump 62 49),((45,49),Jump [50,61]),((45,50),SLiteral (Left (Special '{')) 51),((45,51),Jump [52,55]),((45,53),Jump [54,55]),((45,52),Long_Jump 52 53),((45,54),SLiteral (Left (Special ',')) 52),((45,55),SLiteral (Left (Special '}')) 56),((45,56),SLiteral (Left (Special '{')) 57),((45,58),Jump [59,60]),((45,57),Long_Jump 52 58),((45,59),SLiteral (Left (Special ',')) 57),((45,60),SLiteral (Left (Special '}')) 61),((46,8),End),((46,0),Jump [1,5,7]),((46,1),Jump [2,4]),((46,2),Long_Jump 53 3),((46,3),SLiteral (Left (ReservedOp "<-")) 4),((46,4),Long_Jump 41 8),((46,5),SLiteral (Left (ReservedID "let")) 6),((46,6),Long_Jump 11 8),((46,7),Jump [8]),((47,4),End),((47,0),Jump [1,4]),((47,2),Jump [3,4]),((47,1),Long_Jump 48 2),((47,3),SLiteral (Left (Special ';')) 1),((48,10),End),((48,0),Jump [1,9]),((48,1),Long_Jump 53 2),((48,2),Jump [3,5]),((48,3),SLiteral (Left (ReservedOp "->")) 4),((48,4),Long_Jump 41 6),((48,5),Long_Jump 49 6),((48,6),Jump [7,10]),((48,7),SLiteral (Left (ReservedID "where")) 8),((48,8),Long_Jump 11 10),((48,9),Jump [10]),((49,4),End),((49,3),Jump [0,4]),((49,0),Long_Jump 40 1),((49,1),SLiteral (Left (ReservedOp "->")) 2),((49,2),Long_Jump 41 3),((50,6),End),((50,0),Jump [1,3]),((50,2),Jump [1,3]),((50,1),Long_Jump 51 2),((50,3),Long_Jump 41 4),((50,4),Jump [5,6]),((50,5),SLiteral (Left (Special ';')) 6),((51,10),End),((51,0),Jump [1,6,9]),((51,1),Jump [2,4]),((51,2),Long_Jump 53 3),((51,3),SLiteral (Left (ReservedOp "<-")) 4),((51,4),Long_Jump 41 5),((51,5),SLiteral (Left (Special ';')) 10),((51,6),SLiteral (Left (ReservedID "let")) 7),((51,7),Long_Jump 11 8),((51,8),SLiteral (Left (Special ';')) 10),((51,9),SLiteral (Left (Special ';')) 10),((52,3),End),((52,0),Long_Jump 60 1),((52,1),SLiteral (Left (ReservedOp "=")) 2),((52,2),Long_Jump 41 3),((53,5),End),((53,0),Jump [1,4]),((53,1),Long_Jump 59 2),((53,2),SLiteral (Right (Func "is_plus" (is_plus))) 3),((53,3),SLiteral (Right (Func "is_integer" (is_integer))) 5),((53,4),Long_Jump 54 5),((54,8),End),((54,6),Jump [7,8]),((54,0),Jump [1,5]),((54,1),SLiteral (Right (Func "is_negative" (is_negative))) 2),((54,2),Jump [3,4]),((54,3),SLiteral (Right (Func "is_integer" (is_integer))) 6),((54,4),SLiteral (Right (Func "is_float" (is_float))) 6),((54,5),Long_Jump 55 6),((54,7),Long_Jump 66 0),((55,5),End),((55,0),Jump [1,2]),((55,1),Long_Jump 56 5),((55,2),Long_Jump 58 3),((55,4),Jump [3,5]),((55,3),Long_Jump 56 4),((56,27),End),((56,0),Jump [1,5,6,13,14,15,20,25]),((56,1),Long_Jump 59 2),((56,2),Jump [3,27]),((56,3),SLiteral (Left (ReservedOp "@")) 4),((56,4),Long_Jump 56 27),((56,5),Long_Jump 58 27),((56,6),Long_Jump 62 7),((56,7),SLiteral (Left (Special '{')) 8),((56,8),Jump [9,12]),((56,10),Jump [11,12]),((56,9),Long_Jump 57 10),((56,11),SLiteral (Left (Special ',')) 9),((56,12),SLiteral (Left (Special '}')) 27),((56,13),SLiteral (Right (Func "is_literal" (is_literal))) 27),((56,14),SLiteral (Left (ReservedID "_")) 27),((56,15),SLiteral (Left (Special '(')) 16),((56,17),Jump [18,19]),((56,16),Long_Jump 53 17),((56,18),SLiteral (Left (Special ',')) 16),((56,19),SLiteral (Left (Special ')')) 27),((56,20),SLiteral (Left (Special '[')) 21),((56,22),Jump [23,24]),((56,21),Long_Jump 53 22),((56,23),SLiteral (Left (Special ',')) 21),((56,24),SLiteral (Left (Special ']')) 27),((56,25),SLiteral (Left (ReservedOp "~")) 26),((56,26),Long_Jump 56 27),((57,3),End),((57,0),Long_Jump 60 1),((57,1),SLiteral (Left (ReservedOp "=")) 2),((57,2),Long_Jump 53 3),((58,9),End),((58,0),Jump [1,6,8]),((58,1),SLiteral (Left (Special '(')) 2),((58,2),Jump [3,5]),((58,4),Jump [3,5]),((58,3),SLiteral (Left (Special ',')) 4),((58,5),SLiteral (Left (Special ')')) 9),((58,6),SLiteral (Left (Special '[')) 7),((58,7),SLiteral (Left (Special ']')) 9),((58,8),Long_Jump 62 9),((59,5),End),((59,0),Jump [1,2]),((59,1),SLiteral (Right (Func "is_VarID" (is_VarID))) 5),((59,2),SLiteral (Left (Special '(')) 3),((59,3),SLiteral (Right (Func "is_VarSym" (is_VarSym))) 4),((59,4),SLiteral (Left (Special ')')) 5),((60,5),End),((60,0),Jump [1,2]),((60,1),SLiteral (Right (Func "is_QVarID" (is_QVarID))) 5),((60,2),SLiteral (Left (Special '(')) 3),((60,3),SLiteral (Right (Func "is_QVarSym" (is_QVarSym))) 4),((60,4),SLiteral (Left (Special ')')) 5),((61,5),End),((61,0),Jump [1,2]),((61,1),SLiteral (Right (Func "is_ConID" (is_ConID))) 5),((61,2),SLiteral (Left (Special '(')) 3),((61,3),SLiteral (Right (Func "is_ConSym" (is_ConSym))) 4),((61,4),SLiteral (Left (Special ')')) 5),((62,5),End),((62,0),Jump [1,2]),((62,1),SLiteral (Right (Func "is_QConID" (is_QConID))) 5),((62,2),SLiteral (Left (Special '(')) 3),((62,3),Long_Jump 69 4),((62,4),SLiteral (Left (Special ')')) 5),((63,5),End),((63,0),Jump [1,2]),((63,1),SLiteral (Right (Func "is_VarSym" (is_VarSym))) 5),((63,2),SLiteral (Left (Special '`')) 3),((63,3),SLiteral (Right (Func "is_VarID" (is_VarID))) 4),((63,4),SLiteral (Left (Special '`')) 5),((64,5),End),((64,0),Jump [1,2]),((64,1),SLiteral (Right (Func "is_QVarSym" (is_QVarSym))) 5),((64,2),SLiteral (Left (Special '`')) 3),((64,3),SLiteral (Right (Func "is_QVarID" (is_QVarID))) 4),((64,4),SLiteral (Left (Special '`')) 5),((65,5),End),((65,0),Jump [1,2]),((65,1),SLiteral (Right (Func "is_ConSym" (is_ConSym))) 5),((65,2),SLiteral (Left (Special '`')) 3),((65,3),SLiteral (Right (Func "is_ConID" (is_ConID))) 4),((65,4),SLiteral (Left (Special '`')) 5),((66,5),End),((66,0),Jump [1,2]),((66,1),Long_Jump 69 5),((66,2),SLiteral (Left (Special '`')) 3),((66,3),SLiteral (Right (Func "is_QConID" (is_QConID))) 4),((66,4),SLiteral (Left (Special '`')) 5),((67,3),End),((67,0),Jump [1,2]),((67,1),Long_Jump 63 3),((67,2),Long_Jump 65 3),((68,3),End),((68,0),Jump [1,2]),((68,1),Long_Jump 64 3),((68,2),Long_Jump 66 3),((69,3),End),((69,0),Jump [1,2]),((69,1),SLiteral (Left (ReservedOp ":")) 3),((69,2),SLiteral (Right (Func "is_QConSym" (is_QConSym))) 3)]
